//
//  File:           Triangulation.java
//  Description:    
//  Created:        14.05.97
//  Last Modified:  14.05.97
//  Author:         Bettina Speckmann
//  mail:           speckman@cs.ubc.ca
//

import java.awt.*;
import CompGeom.Point;
import CompGeom.Homog;
import BTriangle;


public class Triangulation {

  public Point P[], testp;
  public int CW[], CCW[];
  public BTriangle bt;

  private int freept = 0, idx, ddraw = 0;

  public Triangulation(int n) {
    
    P = new Point[n];
    CW = new int[3*n];
    CCW = new int[3*n];

    bt = new BTriangle();
    P[0] = bt.p1;
    P[1] = bt.p2;
    P[2] = bt.p3;
    CW[0] = 0;
    CW[1] = 1;
    CW[2] = 2;
    CW[3] = 3;
    CW[4] = 4;
    CW[5] = 5;
    CW[6] = 6;
    CW[7] = 7;
    CW[8] = 8;
    CCW[0] = 0;
    CCW[1] = 6;
    CCW[2] = 2;
    CCW[3] = 3;
    CCW[4] = 4;
    CCW[5] = 1;
    CCW[6] = 5;
    CCW[7] = 7;
    CCW[8] = 8;
    freept = 3;

  }


  // edges are stored counter-clock-wise

  public int Inc(int e) { return(((int)(e/3))*3 + (e+1)%3); }
  
  public int Dec(int e) { return(((int)(e/3))*3 + (e-1)%3); }


  public boolean wedge_ccw(int e) {

    if(CW[CCW[e]] == Inc(e)) return(true);
    else 
      if(CCW[CW[Inc(e)]] == e) return(true);
      else return(false);
  }

  public boolean wedge_cw(int e) {

    if(CCW[CW[e]] == Dec(e)) return(true);
    else 
      if(CW[CCW[Dec(e)]] == e) return(true);
      else return(false);
  } 


  public int next_ccw(int e) {
    
    if(wedge_ccw(e)) return(Inc(e));
    else return(CCW[e]);
  }

  public int next_ccw_inc(int e) {

    if(wedge_ccw(e)) return(CCW[Inc(e)]);
    else return(CCW[e]);
  }


  public int next_cw(int e) {

    if(wedge_cw(e)) return(Dec(e));
    else return(CW[e]);
  }

  public int next_cw_inc(int e) {

    if(wedge_cw(e)) return(CW[Dec(e)]);
    else return(CW[e]);
  }


  public int op_ccw(int e) {
    
    if(wedge_ccw(e))
      if(CW[CCW[e]] == Inc(e)) return(CCW[e]);
      else return(CW[Inc(e)]);
    else
      if(wedge_ccw(CCW[e])) return(Inc(CCW[e])); 
      else return(CCW[CCW[e]]); 
  }

  public int op_cw(int e) {

    if(wedge_cw(e))
      if(CCW[CW[e]] == Dec(e)) return(CW[e]);
      else return(CCW[Dec(e)]);
    else
      if(wedge_cw(CW[e])) return(Dec(CW[e]));
      else return(CW[CW[e]]); 
  }


  // locate a point in the triangulation, return an edge of the inclosing 
  // triangle

  int locate(Point q, int hint) //throws duplicatePointExeption 
    {
    int e, left, right, i = 3*freept;
    Point a, b, c;
    
    e = hint;

    do{
      if(Point.left(Org(e), Dest(e), q)) {
	a = Org(e);
	b = Dest(e);
	if(wedge_ccw(e)) c = Dest(Inc(e));
	else c = Org(CCW[e]);
	left = next_ccw(e);
	right = op_ccw(e);
      }
      else {
	a = Dest(e);
	b = Org(e);
	if(wedge_cw(e)) c = Dest(Dec(e));
	else c = Org(CW[e]);
	left = op_cw(e);
	right = next_cw(e);
      }
      
      //if(q.equals(a) || q.equals(b) || q.equals(c)) 
      //throw new duplicatePointException(e);
      
      if(Point.right(a, c, q)) {
	if(Point.right(c, b, q)) return(e);
	else e = right;
      }
      else e = left;

    } while(i-- > 0);
    return (-1); // we should never get here
  }


  // add point p in triangle on "side" of edge e
  // side == 0 => left, side == 1 => right

  void addPt(Point p, int e, int side) {
    int help;
    
    P[freept] = p;
    idx = 3*freept + e%3;
    freept++;
    if(side == 1) { 
      if(wedge_cw(e)) CCW[Inc(idx)] = Inc(CW[e]); 
      else CCW[Inc(idx)] = CW[e];
      if(wedge_cw(e)) CCW[Dec(idx)] = CCW[Dec(e)];
      else if(wedge_cw(CW[e])) CCW[Dec(idx)] = Inc(CW[CW[e]]);
      else CCW[Dec(idx)] = CW[CW[e]];
      CCW[idx] = e;
      CW[idx] = Dec(CCW[Inc(idx)]);
      CW[Inc(idx)] = Dec(CCW[Dec(idx)]);
      CW[Dec(idx)] = Dec(e);
      
      help = op_cw(e); 
      if(wedge_cw(e)) CCW[Dec(e)] = Inc(idx);
      else CW[CW[e]] = Dec(idx);
      if(CW[help] == e) CW[help] = idx;
      else CCW[help] = Dec(idx);
      CW[e] = Inc(idx);
    }
    else {
      if(wedge_ccw(e)) CW[Dec(idx)] = Dec(CCW[e]); 
      else CW[Dec(idx)] = CCW[e];
      if(wedge_ccw(e)) CW[Inc(idx)] = CW[Inc(e)];
      else if(wedge_ccw(CCW[e])) CW[Inc(idx)] = Dec(CCW[CCW[e]]);
      else CW[Inc(idx)] = CCW[CCW[e]];
      CW[idx] = e;
      CCW[idx] = Inc(CW[Dec(idx)]);
      CCW[Dec(idx)] = Inc(CW[Inc(idx)]);
      CCW[Inc(idx)] = Inc(e);  
     
      help = op_ccw(e);
      if(wedge_ccw(e)) CW[Inc(e)] = Dec(idx);
      else CCW[CCW[e]] = Inc(idx);
      if(CCW[help] == e) CCW[help] = idx;
      else CW[help] = Inc(idx);
      CCW[e] = Dec(idx);
    } 
  }
  

  public void insert(Point p) {
    int e;

    e = locate(p, 1);
    if(Point.right(Org(e), Dest(e), p)) addPt(p, e, 1);
    else addPt(p, e, 0);

    audit(idx); audit(Dec(idx)); audit(Inc(idx));
    restoreDelaunay(idx, freept-1); // EVIL use of global idx
  }


  // everything needed to do Delaunay Triangulations


  boolean DelaunayEdge(int e) {
    Point a, b, c, d;

    // catch the bounding triangle edges
    
    if((e == 1) || (e == 5) || (e == 6)) return(true);

    else {

      if(wedge_ccw(e)) a = Dest(Inc(e));
      else a = Org(CCW[e]);
      b = Org(e);
      if(wedge_cw(e)) c = Dest(Dec(e));
      else c = Org(CW[e]);
      d = Dest(e);
      
      // points have to be in clockwise order 
      // (strangely enough, could be due to the flipped coordinates)

      return(!a.inCircle(d, c, b));
    }
  }

  
  // v is the new inserted vertex, needed to determine the direction of e

  final static int MAXI = 10;	// iteration limit
  final static int MAXGUARD = 100; // iteration limit
  void restoreDelaunay(int e, int v) {
    int next, d, i, guard = MAXGUARD;
    
    // first remember the destination of your starting edge
    
    d = dest(e);

    do {
      if(((int)e/3) == v) next = op_ccw(e);
      else next = next_cw(e);

      i = MAXI;
      while(!DelaunayEdge(next) && (--i > 0)) {
	e = swap(next, e, v);
      	if(((int)e/3) == v) next = op_ccw(e);
      	else next = next_cw(e);
      }
      if (i <= 0) System.out.println("ERROR in restore loop "+e+","+next);

      if(org(e) == v) e = next_ccw(e);
      else e = op_cw(e);
      
    } while ((dest(e) != d) && (org(e) != d) && (--guard > 0));

    if (guard <= 0) System.out.println("ERROR in restore Delaunay "+e);
    audit();
  }
  

  
  // return new name of e
  
  int swap(int next, int e, int v) {
    int conf, help, help2, help3, help4, d;

    // first find out how e and next lay to each other and to v
    
    if(org(e) == v) {
      if(next_ccw(next) == e) conf = 1;
      else conf = 2;
    }
    else{
      if(next_ccw(next) == e) conf = 3;
      else conf = 4;
    }
    
    ddraw = next;

    switch (conf) {
    case 1: {
      
      System.out.println("case 1");

      help = op_cw(e);
      help2 = op_cw(next);
      help4 = next_ccw(e);
      d = dest(next);

      if(org(help) == org(next)) {
	if(wedge_cw(e)) {
	  CCW[Dec(e)] = next;
	  CCW[next] = Inc(CW[e]);
	}
	else CCW[next] = CW[e];
	CW[help] = Dec(e);
      }
      else {
	CCW[next] = help;
	if(wedge_cw(e)) CCW[Dec(e)] = next;
      }
      
      CW[e] = next;
      help3 = next_cw(next);

      if(org(help3) == org(next)) CCW[help3] = e;
      else CW[help3] = e;
     
      if(org(help4) == org(e)) {

	if(wedge_cw(next)) CW[help4] = CW[next];
	else CW[help4] = Dec(help3);

	if(org(help2) == d) CW[help2] = help4;
	else CCW[help2] = e;
      }
      else{
	if(org(help2) == d) {
	  CW[help2] = e;
	  if(wedge_cw(next)) CCW[help4] = Inc(CW[next]);
	  else CCW[help4] = help3;
	}
	else {
	  CCW[help4] = help2;
	  CCW[help2] = e;
	}
      }

    audit(e); audit(next);
      return(next); 
    }
    case 2: {

      System.out.println("case 2");

      if(wedge_ccw(e)) {
	help = Inc(e);
	help2 = next_ccw(next);

	CCW[help] = next;
	
	
      }

    audit(e); audit(next);
      return(e);
    }
    default:
       System.out.println("default");
       if(org(e) == v) return(next_ccw(e));
       else return(op_cw(e));
    }
  }
 


  // reset

  public void clearcounts() {

    CW[0] = 0;
    CW[1] = 1;
    CW[2] = 2;
    CW[3] = 3;
    CW[4] = 4;
    CW[5] = 5;
    CW[6] = 6;
    CW[7] = 7;
    CW[8] = 8;
    CCW[0] = 0;
    CCW[1] = 6;
    CCW[2] = 2;
    CCW[3] = 3;
    CCW[4] = 4;
    CCW[5] = 1;
    CCW[6] = 5;
    CCW[7] = 7;
    CCW[8] = 8;
    freept = 3;

    ddraw = 0;
  }
    

  
  
  public void assert(boolean test, String s) {
    if (!test) System.out.println("ASSERT FAILED: "+s);
  }

  int one(boolean f) { return f ? 1 : 0; }

  public void audit(int e) {
    boolean pass = true;
    if (wedge_cw(e)) {
      if ((e-CW[e])%3 != 0) { 
	System.out.println("Bad wedge CW["+e+","+(e%3)
			   +"]="+CW[e]+","+(CW[e]%3));
	pass = false;
      }
    } else {
      if ((Inc(e)-CW[e])%3 != 0) { 
	System.out.println("Bad non-wedge CW["+e+","+(Inc(e)%3)
			   +"]="+CW[e]+","+(CW[e]%3));
	pass = false;
      }
    }
    if (wedge_ccw(e)) {
      if ((e-CCW[e])%3 != 0) { 
	System.out.println("Bad wedge CCW["+e+","+(e%3)
			   +"]="+CCW[e]+","+(CCW[e]%3));
	pass = false;
      }
    } else {
      if ((Dec(e)-CCW[e])%3 != 0) { 
	System.out.println("Bad non-wedge CCW["+e+","+(Dec(e)%3)
			   +"]="+CCW[e]+","+(CCW[e]%3));
	pass = false;
      }
    }
    if (!pass) System.out.println("e ccw cw inc dec\n"
   +e+" "+CCW[e]+" "+CW[e]+" "+Inc(e)+" "+Dec(e)+"\n"
   +Inc(e)+" "+CCW[Inc(e)]+" "+CW[Inc(e)]+" "+Inc(Inc(e))+" "+Dec(Inc(e))+"\n"
   +Dec(e)+" "+CCW[Dec(e)]+" "+CW[Dec(e)]+" "+Inc(Dec(e))+" "+Dec(Dec(e))+"\n"
   +CCW[e]+" "+CCW[CCW[e]]+" "+CW[CCW[e]]+" "+Inc(CCW[e])+" "+Dec(CCW[e])+"\n"
   +CW[e]+" "+CCW[CW[e]]+" "+CW[CW[e]]+" "+Inc(CW[e])+" "+Dec(CW[e]));
  }

  public void audit() { for (int i = 9; i < 3*freept; i++) audit(i); }








  // all stuff for drawing the triangulation
  
  
  
  final int org(int e) { return((int)e/3);}
  final int dest(int e) {
    return (wedge_ccw(e) ? org(CW[Inc(e)]) : org(CCW[CCW[e]])); 
  }

  public Point Org(int e) { return(P[org(e)]); } 
  public Point Dest(int e) { return(P[dest(e)]); }
  
  public void draw_arrow(Graphics g, int e) {
      Point p = Org(e);
      Point q = Dest(e);
      draw_arrow(g, p, q);
      Point l = Point.lincomb(.6, p, .4, q);
      g.drawString(Integer.toString(e), (int)l.x, -(int)l.y);
    }

  final static int arrowlength = 10;
  final static int arrowhalfwidth = 4;
    
  public void draw_arrow(Graphics g, Point p, Point q) {
    Point l, r, s, t; 
    int xPoints[], yPoints[];
    
    xPoints = new int[3];
    yPoints = new int[3];
    
    l = Point.sub(q, p);
    s = Point.lincomb(p, 0.5, l);
    l.normalize();
    t = Point.lincomb(s, -arrowlength, l);
    l = l.perp();
    r = Point.lincomb(t, -arrowhalfwidth, l);
    t = Point.lincomb(t, arrowhalfwidth, l);
    xPoints[0] = (int)r.x;
    xPoints[1] = (int)s.x;
    xPoints[2] = (int)t.x;
    yPoints[0] = (int)(-r.y);
    yPoints[1] = (int)(-s.y);
    yPoints[2] = (int)(-t.y);
    g.drawLine((int)p.x, (int)(-p.y), (int)q.x, (int)(-q.y));
    g.fillPolygon(xPoints, yPoints, 3);
  }


  public void delaunay_draw(Graphics g, int e) {
      Point p = Org(e);
      Point q = Dest(e);
      g.setColor(Color.orange);
      g.drawLine((int)p.x, (int)(-p.y), (int)q.x, (int)(-q.y));
    //    draw_arrow(g, e);
  }


  public void draw(Graphics g) {
    int i;
    
    g.setColor(Color.blue);
    draw_arrow(g, 6);
    for(i = 9; i < 3*freept; i+=3) {
      draw_arrow(g, i);
    }
    g.setColor(Color.green);
    draw_arrow(g, 1);
    for(i = 10; i < 3*freept; i+=3) {
      draw_arrow(g, i);
    }
    g.setColor(Color.red);
    draw_arrow(g, 5);
    for(i = 11; i < 3*freept; i+=3) {
     draw_arrow(g, i);
    }
    g.setColor(Color.black);
    for(i = 0; i < freept; i++) {
      g.fillOval((int)P[i].x-3, (int)-P[i].y-3, 6, 6);
    }
    
    if(ddraw != 0) delaunay_draw(g, ddraw);

  }

}
